import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Prepare the rule documentation file from the rulesets found in he classpath of PMD.
 * @author KoR
 */
public class BuildRuleDoc implements Comparable
{
   /**
    * Recognize a rule in the ruleset AND the rule file itself by the start.
    */
   private static final String RULE_START = "<rule ";
   /**
    * Insert the created content AFTER this content.
    */
   private static final String TEMPLATE_INSERTER = "<!-- CONTENT -->";

   // instance of the rule
   private final String rulePath;
   private final String ruleSet;
   private final String ruleName;
   private String ruleMsg;
   private String ruleDesc;
   private String author;
   private String ruleExample;

   public BuildRuleDoc(String pLine)
   {
      rulePath = pLine.substring(0, pLine.indexOf('/'));
      pLine = pLine.substring(pLine.indexOf('/') + 1);
      ruleSet = pLine.substring(0, pLine.indexOf('/'));
      pLine = pLine.substring(pLine.indexOf('/') + 1);
      ruleName = pLine;
   }

   public int compareTo(Object pO)
   {
      return ruleName.compareTo(((BuildRuleDoc) pO).ruleName);
   }

   public String toString()
   {
      return this.getClass().getName() + "{ruleSet=" + this.ruleSet + ",ruleName=" + this.ruleName + '}';
   }

   private String extractContent(String pContent, String pFrom, String pTo)
   {
      int iMsgStart = pContent.indexOf(pFrom);
      if (iMsgStart == -1)
      {
         throw new IllegalArgumentException("value " + pFrom + " not found in input"); //  + pContent
      }
      iMsgStart += pFrom.length();
      final int iMsgEnd = pContent.indexOf(pTo, iMsgStart);
      String res = pContent.substring(iMsgStart, iMsgEnd).trim();
      if (res.startsWith("<![CDATA["))
      {
         res = res.substring(9);
      }
      if (res.endsWith("]]>"))
      {
         res = res.substring(0, res.length() - 3);
      }
      return res.trim();
   }

   private static String readFullyFrom(BufferedReader pIn) throws IOException
   {
      final char[] buffer = new char[1024 * 1024];
      final int size = pIn.read(buffer);
      return new String(buffer, 0, size);
   }

   private static String replace(String pIn, String pFrom, String pTo)
   {
      String res = pIn;
      int i = res.indexOf(pFrom);
      while (i != -1)
      {
         res = res.substring(0, i) + pTo + res.substring(i + pFrom.length());
         i = res.indexOf(pFrom);
      }
      return res;
   }

   public void loadMessages() throws IOException
   {
      // load rule from classpath
      final InputStream inStream = BuildRuleDoc.class.getClassLoader().getResourceAsStream(rulePath + '/' + ruleSet);
      if (inStream == null)
      {
         throw new IllegalArgumentException("ruleset " + ruleSet + " not found in classpath");
      }

      final BufferedReader in = new BufferedReader(new InputStreamReader(inStream));
      String rulesetxml = readFullyFrom(in);
      in.close();

      // convert "wrong" format to allow code further down to function properly
      rulesetxml = replace(rulesetxml, "<rule\n", "<rule ");
      rulesetxml = replace(rulesetxml, "<rule\t", "<rule ");
      rulesetxml = replace(rulesetxml, "<rule  ", "<rule ");

      // parse for the rule data
      rulesetxml = extractContent(rulesetxml, "<rule name=\"" + ruleName + "\"", "</rule>");
      ruleMsg = extractContent(rulesetxml, "message=\"", "\"");
      ruleDesc = "";
      if (rulesetxml.indexOf("<description>") != -1)
      {
         ruleDesc = extractContent(rulesetxml, "<description>", "</description>");
         if (ruleDesc.equals(ruleMsg))
         {
            ruleDesc = "";
         }
      }
      else
      {
         System.out.println("value " + "<description>" + " not found in rule " + ruleName);
      }
      author = "PMD";
      if (rulesetxml.indexOf("<author>") != -1)
      {
         author = extractContent(rulesetxml, "<author>", "</author>");
      }
      ruleExample = extractContent(rulesetxml, "<example>", "</example>");
   }

   public String toHtml()
   {
      final StringBuffer buf = new StringBuffer();
      buf.append("<fieldset>");
      buf.append(System.getProperty("line.separator"));
      buf.append("<legend><span><a name=\"");
      buf.append(ruleName);
      buf.append("\">");
      buf.append(ruleName);
      buf.append("</a></span> (");
      buf.append(ruleSet);
      buf.append("):</legend>");
      buf.append(System.getProperty("line.separator"));
      buf.append("<div class=\"message\">");
      buf.append(ruleMsg);
      buf.append("</div>");
      if (ruleDesc.length() > 0)
      {
         buf.append("<br />");
         buf.append(System.getProperty("line.separator"));
         buf.append("<div class=\"description\">");
         buf.append(ruleDesc);
         buf.append("</div>");
      }
      if (ruleExample.length() > 0)
      {
         buf.append(System.getProperty("line.separator"));
         buf.append("<pre>");
         buf.append(System.getProperty("line.separator"));
         buf.append(ruleExample);
         buf.append(System.getProperty("line.separator"));
         buf.append("</pre>");
      }

      buf.append("<div class=\"author\">");
      buf.append(author);
      buf.append("</div>");

      buf.append("</fieldset>");
      buf.append(System.getProperty("line.separator"));
      buf.append(System.getProperty("line.separator"));
      return buf.toString();
   }

   // static helper

   private static BuildRuleDoc getRule(String pLine) throws IOException
   {
      //   line is <rule ref="rulesets/basic.xml/DoubleCheckedLockingRule"/>
      pLine = pLine.trim();
      if (!pLine.startsWith(RULE_START))
      {
         throw new IllegalArgumentException("input not starting with " + RULE_START + ": " + pLine);
      }
      // skip start and end
      pLine = pLine.substring(pLine.indexOf('\"') + 1);
      pLine = pLine.substring(0, pLine.indexOf('\"'));

      // parse rule properties from input
      final BuildRuleDoc rule = new BuildRuleDoc(pLine);
      rule.loadMessages();
      return rule;
   }

   private static void loadAllRules(List pList, String pRuleSetName) throws IOException
   {
      final BufferedReader in = new BufferedReader(new FileReader(pRuleSetName));
      String t = null;
      while ((t = in.readLine()) != null)
      {
         t = t.trim();
         if (t.startsWith(RULE_START))
         {
            pList.add(getRule(t));
         }
      }
      in.close();
      Collections.sort(pList);
   }

   private static final void output(String pMessage)
   {
      System.out.println(pMessage);
   }

   /**
    * @param Names of Rule sets ending with .xml and then two html template and target file.
    */
   public static void main(String[] pArgs)
   {
      try
      {
         // load rules
         final List rules = new ArrayList();
         int j = 0;
         while (pArgs.length > j && pArgs[j].endsWith(".xml"))
         {
            final String ruleset = pArgs[j]; // contains name of rule set file 
            if (!new File(ruleset).exists())
            {
               throw new IOException("file " + ruleset + " does not exist");
            }
            loadAllRules(rules, ruleset);
            j++;
         }

         // load template
         final BufferedReader in = new BufferedReader(new FileReader(pArgs[j++])); // template html
         String termplate = readFullyFrom(in);
         in.close();
         final int splitIndex = termplate.indexOf(TEMPLATE_INSERTER) + TEMPLATE_INSERTER.length();

         // insert content
         final StringBuffer buf = new StringBuffer();
         buf.append(termplate.substring(0, splitIndex));
         buf.append(System.getProperty("line.separator"));
         for (Iterator i = rules.iterator(); i.hasNext();)
         {
            buf.append(((BuildRuleDoc) i.next()).toHtml());
         }
         buf.append(termplate.substring(splitIndex));

         // write file
         final String saveAs = pArgs[j++]; // contains name of report html
         final FileWriter out = new FileWriter(saveAs);
         out.write(buf.toString());
         out.flush();
         out.close();
         output("Rule description created.");

      }
      catch (IOException ex)
      {
         output("Error: " + ex);
         ex.printStackTrace();
      }
   }

}
